import 'dart:async';

import 'package:bloc/bloc.dart';
import 'package:flutter_secure_storage/flutter_secure_storage.dart';

import '../cidaas_login_provider.dart';
import './../entity/token_entity.dart';
import 'authentication_storage_helper.dart';

part 'authentication_event.dart';
part 'authentication_state.dart';

/// The authentication Bloc used to determine which screen should get displayed
class AuthenticationBloc
    extends Bloc<AuthenticationEvent, AuthenticationState> {
  static AuthenticationBloc? _instance;
  static String reRouteToAfterLogin = '';

  static AuthStorageHelper? _authStorageHelper;

  /// Factory returns Singleton AuthenticationBloc
  ///
  /// if [storageHelper] is given upon first creation, uses this AuthStorageHelper
  /// if not, creates a new one
  factory AuthenticationBloc({FlutterSecureStorage? secureStorage}) {
    if (AuthenticationBloc._instance != null) {
      return AuthenticationBloc._instance!;
    } else {
      return AuthenticationBloc._internal(secureStorage: secureStorage);
    }
  }

  /// Internal constr.
  AuthenticationBloc._internal({FlutterSecureStorage? secureStorage})
      : super(AuthenticationHasLoggedOutState()) {
    if (secureStorage == null) {
      AuthenticationBloc._authStorageHelper = AuthStorageHelper();
    } else {
      AuthenticationBloc._authStorageHelper =
          AuthStorageHelper(storage: secureStorage);
    }

    on<AuthenticationStartedEvent>(startEvent);
    on<AuthenticationLoggedInEvent>(loggedInEvent);
    on<AuthenticationLoggedOutEvent>(loggedOutEvent);
  }

  Future<void> startEvent(AuthenticationStartedEvent event,
      Emitter<AuthenticationState> emit) async {
    emit(AuthenticationInProgressState());
    try {
      if (await CidaasLoginProvider.isAuth()) {
        emit(AuthenticationSuccessState(
            tokenEntity: await _authStorageHelper!.getCurrentToken()));
      } else {
        emit(AuthenticationShowLoginWithBrowserState());
      }
    } catch (e) {
      emit(AuthenticationFailureState(error: e.toString()));
    }
  }

  Future<void> loggedInEvent(AuthenticationLoggedInEvent event,
      Emitter<AuthenticationState> emit) async {
    emit(AuthenticationInProgressState());
    if (event.tokenEntity == null || event.tokenEntity!.accessToken == null) {
      emit(const AuthenticationFailureState(
          error: 'No access token received after login'));
    } else {
      await _authStorageHelper?.persistTokenEntity(event.tokenEntity!);
      emit(AuthenticationSuccessState(tokenEntity: event.tokenEntity!));
    }
  }

  Future<void> loggedOutEvent(AuthenticationLoggedOutEvent event,
      Emitter<AuthenticationState> emit) async {
    emit(AuthenticationInProgressState());
    await _authStorageHelper?.deleteToken();
    emit(AuthenticationHasLoggedOutState());
  }

  @override
  Future<void> close() async {
    _instance = null;
    _authStorageHelper?.close();
    _authStorageHelper = null;
    super.close();
  }
}
